/*
 * FILE NAME board_setup.c
 *
 * BRIEF MODULE DESCRIPTION
 *  Simple setup for different board configurations.
 *
 *  Author: CradlePoint Technology, Inc.  <source@cradlepoint.com>
 *          Steve Johnson <stevej@cradlepoint.com>
 *
 * Copyright 2012, CradlePoint Technology, Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2
 *  as published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to:
 *  Free Software Foundation
 *  51 Franklin Street, Fifth Floor
 *  Boston, MA  02111-1301  USA
 */

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/jiffies.h>
#include <linux/timer.h>
#include <linux/delay.h>
#include <linux/ioport.h>
#include <asm/mach-ralink/cpboard.h>
#include <asm/io.h>
#include <linux/platform_device.h>
#include <linux/mtd/partitions.h>
#include <linux/mtd/physmap.h>
#include "linux/module.h"
#include "board_setup.h"

void __iomem *base_addr;

static void twiddle_leds(unsigned long unused);
DEFINE_TIMER(led_timer, twiddle_leds, 0, 0);

static void reset_pin(int pin)
{

	if (pin < 24) {
		WRITE_PIO((1 << pin), RALINK_REG_PIORESET);
	} else if (pin < 40) {
		pin -= 24;
		WRITE_PIO((1 << pin), RALINK_REG_PIO3924RESET);
	} else if (pin < 72) {
		pin -= 40;
		WRITE_PIO((1 << pin), RALINK_REG_PIO7140RESET);
	} else {
		pin -= 72;
		WRITE_PIO((1 << pin), RALINK_REG_PIO9572RESET);
	}

}

static void set_pin(int pin)
{

	if (pin < 24) {
		WRITE_PIO((1 << pin), RALINK_REG_PIOSET);
	} else if (pin < 40) {
		pin -= 24;
		WRITE_PIO((1 << pin), RALINK_REG_PIO3924SET);
	} else if (pin < 72) {
		pin -= 40;
		WRITE_PIO((1 << pin), RALINK_REG_PIO7140SET);
	} else {
		pin -= 72;
		WRITE_PIO((1 << pin), RALINK_REG_PIO9572SET);
	}

}

static void __init reset_gpios(void)
{
	int i;

	//  The GPIO driver hasn't been installed yet, so
	// do some basic pin setup here.
	for (i=0; i < ARRAY_SIZE(pins_to_reset); i++) {
		reset_pin(pins_to_reset[i]);	
	}

	for (i=0; i < ARRAY_SIZE(pins_to_set); i++) {
		set_pin(pins_to_set[i]);
	}
}

static void __init check_congo_internal_modem(void)
{
	// Check for an internal modem on GPIO 64
	int pin64 = 0x1000000;

	// Set pin to input
	WRITE_PIO((GPIO_OUTPUT_PIN_MASK_40_71 & ~(pin64)), RALINK_REG_PIO7140DIR);
	mdelay(10);
	//printk("Dir 0x%x\n", READ_PIO(RALINK_REG_PIO7140DIR));

	// Read input
	pin64 &= READ_PIO(RALINK_REG_PIO7140DATA);
	//printk("Data 0x%x\n", READ_PIO(RALINK_REG_PIO7140DATA));

	if (pin64) {
		// It's a 1?  There's a modem there, reset as an output
		mdelay(10);
		WRITE_PIO(GPIO_OUTPUT_PIN_MASK_40_71, RALINK_REG_PIO7140DIR);  		
		//printk("Dir 0x%x\n", READ_PIO(RALINK_REG_PIO7140DIR));
		//printk("Found internal modem\n");
	} else {
		// It's a 0?  No modem there, leave as an input
		//printk("Found no internal modem\n");
	}

}

/*
  Per-platform, light the Signal Strength LEDs
  in order during boot.
*/
static void twiddle_leds(unsigned long unused)
{
	static int pin_index = 0;

	base_addr = ioremap(RALINK_SYSCTL_ADDR, 0x1000);
	if (pin_index < ARRAY_SIZE(leds_to_twiddle)) {
		reset_pin(leds_to_twiddle[pin_index]);
		pin_index++;
		mod_timer(&led_timer, jiffies + DELAY_TIME);
	}
	iounmap(base_addr);
}

static int __init board_gpio_setup(void)
{
	base_addr = ioremap(RALINK_SYSCTL_ADDR, 0x1000);

	printk("Board GPIO Setup\n");

	WRITE_PIO(0, RALINK_REG_PIOPOL);

	// Set GPIO mode on those products that don't do that completely
	//  in the bootloader
#if defined(PRODUCT_RUBY)
	int reg = READ_PIO(RALINK_REG_GPIOMODE);
	reg &= ~(1 << 9); // set RGMII to Normal mode
	reg &= ~(3 << 7); // set MDIO to Normal mode
	WRITE_PIO(reg, RALINK_REG_GPIOMODE);
#endif
#if defined(PRODUCT_RUBY) || defined(PRODUCT_IBR200) || defined (PRODUCT_CBA550)
	// configure Master Reset Bit to off before enabling
	WRITE_PIO(0x1000, RALINK_REG_PIO3924SET);
#endif

	// Set GPIO direction
	WRITE_PIO(GPIO_OUTPUT_PIN_MASK, RALINK_REG_PIODIR);  
	WRITE_PIO(GPIO_INT_PIN_MASK, RALINK_REG_PIORENA);
	WRITE_PIO(GPIO_INT_FEDGE_PIN_MASK, RALINK_REG_PIOFENA);
	WRITE_PIO(GPIO_POL_MASK, RALINK_REG_PIOPOL);

	WRITE_PIO(GPIO_OUTPUT_PIN_MASK_24_39, RALINK_REG_PIO3924DIR);  
	WRITE_PIO(GPIO_INT_PIN_MASK_24_39, RALINK_REG_PIO3924RENA);
	WRITE_PIO(GPIO_INT_FEDGE_PIN_MASK_24_39, RALINK_REG_PIO3924FENA);
	WRITE_PIO(GPIO_POL_MASK_24_39, RALINK_REG_PIO3924POL);

	WRITE_PIO(GPIO_OUTPUT_PIN_MASK_40_71, RALINK_REG_PIO7140DIR);  
	WRITE_PIO(GPIO_INT_PIN_MASK_40_71, RALINK_REG_PIO7140RENA);
	WRITE_PIO(GPIO_INT_FEDGE_PIN_MASK_40_71, RALINK_REG_PIO7140FENA);
	WRITE_PIO(GPIO_POL_MASK_40_71, RALINK_REG_PIO7140POL);

	if (GPIO_PINS > 72) {
		WRITE_PIO(GPIO_OUTPUT_PIN_MASK_72_95, RALINK_REG_PIO9572DIR);  
		WRITE_PIO(GPIO_INT_PIN_MASK_72_95, RALINK_REG_PIO9572RENA);
		WRITE_PIO(GPIO_INT_FEDGE_PIN_MASK_72_95, RALINK_REG_PIO9572FENA);
		WRITE_PIO(GPIO_POL_MASK_72_95, RALINK_REG_PIO9572POL);
	}

	// Reset the GPIO interrupt registers
	WRITE_PIO(0xffffff, RALINK_REG_PIOINT);
	WRITE_PIO(0xffffff, RALINK_REG_PIOEDGE);
	WRITE_PIO(0xffffff, RALINK_REG_PIO3924INT);
	WRITE_PIO(0xffffff, RALINK_REG_PIO3924EDGE);
	WRITE_PIO(0xffffffff, RALINK_REG_PIO7140INT);
	WRITE_PIO(0xffffffff, RALINK_REG_PIO7140EDGE);
	if (GPIO_PINS > 72) {
		WRITE_PIO(0xffffff, RALINK_REG_PIO9572INT);
		WRITE_PIO(0xffffff, RALINK_REG_PIO9572EDGE);
	}

	reset_gpios();

#if defined(PRODUCT_CONGO) || defined(PRODUCT_RUBY)
	check_congo_internal_modem();
#endif

	// Start the LED timer
	mod_timer(&led_timer, jiffies + DELAY_TIME);

	iounmap(base_addr);

	return 0;
}

device_initcall(board_gpio_setup);

int rt3xxx_check_memo_reg(int key)
{
#define NO_SECURITY_MAGIC	0x27051958
#define SERIAL_MAGIC        0x100000

// These registers may be overwritten.  Keep the value around in case
//  it is used later.  Bitmask 1 == security, 2 = serial
	static int keyvalue;
	int ret = 0;

	base_addr = ioremap(RALINK_SYSCTL_ADDR, 0x1000);

	//printk("%s\n", __FUNCTION__);

	switch (key) {
	case NO_SECURITY:
		if ((NO_SECURITY_MAGIC == READ_PIO(RALINK_SYSCTL_MEMO0)) ||
			(keyvalue & 1)) {
			keyvalue |= 1;
			ret = 1;
		}
		break;
	case SERIAL_CONSOLE:
		if ((SERIAL_MAGIC & READ_PIO(RALINK_SYSCTL_MEMO1)) ||
			(keyvalue & 2)) {
			keyvalue |= 2;
			ret = 1;
		}
		break;

	default:
		break;
	}

	iounmap(base_addr);
	return ret;
}

// Flash mapping:
#if 1
static struct mtd_partition ralink_partitions[] = {
        {
                name:           "Bootloader",  /* mtdblock0 */
                size:           MTD_BOOT_PART_SIZE,  /* 192K */
                offset:         0,
        }, {
                name:           "Config", /* mtdblock1 */
                size:           MTD_CONFIG_PART_SIZE,  /* 64K */
                offset:         MTDPART_OFS_APPEND
        }, {
                name:           "Factory", /* mtdblock2 */
                size:           MTD_FACTORY_PART_SIZE,  /* 64K */
                offset:         MTDPART_OFS_APPEND
#if !defined (CONFIG_RALINK_NAND_BOOT)
        }, {
                name:           "Kernel", /* mtdblock3 */
                size:           0,
                offset:         MTDPART_OFS_APPEND,
        }, {
                name:           "RootFS", /* mtdblock4 */
                size:           0,
                offset:         MTDPART_OFS_APPEND,
#endif
        }, {
                name:           "2nd Filemanager",
                size:           MTD_FILEMGR_PART_SIZE,
                offset:         MTD_PART_SIZE - (MTD_FILEMGR_PART_SIZE * 2),
        }, {
                name:           "Filemanager",
                size:           MTD_FILEMGR_PART_SIZE,
                offset:         MTD_PART_SIZE - MTD_FILEMGR_PART_SIZE,
		}
};

static struct physmap_flash_data physmap_flash_data = {
	.width		= 2,
	.nr_parts   = ARRAY_SIZE(ralink_partitions),
	.parts		= ralink_partitions,
};

#ifdef CONFIG_MTD_PHYSMAP_START
static struct resource physmap_flash_resource = {
	.start		= CONFIG_MTD_PHYSMAP_START,
	.end		= CONFIG_MTD_PHYSMAP_START + CONFIG_MTD_PHYSMAP_LEN - 1,
	.flags		= IORESOURCE_MEM,
};
#endif

static struct platform_device physmap_flash = {
	.name	= "ralink-flash",
	.id		= 0,
	.dev	= {
		.platform_data	= &physmap_flash_data,
	},
#ifdef CONFIG_MTD_PHYSMAP_START
	.num_resources	= 1,
	.resource = &physmap_flash_resource,
#endif
};

#ifdef CONFIG_MTD_PARTITIONS
void physmap_set_partitions(struct mtd_partition *parts, int num_parts)
{
	physmap_flash_data.nr_parts = num_parts;
	physmap_flash_data.parts = parts;
}
#endif

static int __init physmap_init(void)
{
#if defined (CONFIG_MTD_NOR_RALINK) || defined (CONFIG_MTD_ANY_RALINK) || defined (CONFIG_MTD_SPI_RALINK)
	int err;

	//printk("%s init\n", __FUNCTION__);

	err = platform_device_register(&physmap_flash);

	return err;
#endif
}

module_init(physmap_init);
#endif
